import gc
import glob
import json
import os
from pathlib import Path

import click
import torch
import torch.nn as nn
import torch.nn.functional as f
import torch.optim as optim
import wandb
from dotenv import find_dotenv, load_dotenv
from numpy import mean
from sklearn.model_selection import KFold
from torch.utils.data import DataLoader
from tqdm import tqdm

from src.data.camembert_ner_dataset import CamemBertNerDataset
from src.models.classifiers import classifier_factory
from src.models.utils import init_tokenizer, save_model, monitor_output_example, \
    monitor_model_eval, init_model, logger


# pylint: disable=too-many-arguments, too-many-locals
def compute_loss(predictions, targets, criterion=None):
    rearranged_output = predictions.contiguous().view(
        predictions.shape[0] * predictions.shape[1], -1)
    rearranged_target = targets.contiguous().view(-1)
    loss = criterion(rearranged_output, rearranged_target)
    return loss


def train_model(model, classifier, dataloader, optimizer, criterion, device, id_fold, epochs):
    model.train()
    classifier.train()
    epoch_loss = []
    logger.info('🎓 model training step is started')
    for epoch in tqdm(range(epochs), position=0,leave=False, desc='epoch'):
        logger.info("Starting epoch {}".format(epoch + 1))
        for batch in tqdm(dataloader, position=1, leave=False, desc='learning batch'):
            optimizer.zero_grad()

            predictions = classifier(
                model(input_ids=batch['tokens'].to(device),
                      attention_mask=batch['attention_mask'].to(device)).last_hidden_state
            )
            predictions_scores = f.log_softmax(predictions, dim=2)
            loss = compute_loss(predictions_scores, batch['target'].to(device), criterion)
            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), 1.0)
            optimizer.step()
            wandb.log({"loss {}".format(id_fold): loss.item()})
            epoch_loss.append(loss.item())

    logger.info('train epoch loss : {}'.format(mean(epoch_loss)))
    logger.info('🎓 model training step is finished')


def eval_model(model, classifier, dataloader, criterion, device, labels):
    model.eval()
    classifier.eval()
    epoch_loss = []
    evaluation_measure = {
        'recall': [],
        'precision': [],
        'F1 score': []
    }
    logger.info('model evaluation step is started')
    with torch.no_grad():
        batch = None
        for batch in dataloader:
            predictions = classifier(
                model(input_ids=batch['tokens'].to(device),
                      attention_mask=batch['attention_mask'].to(device)).last_hidden_state
            )
            predictions_scores = f.log_softmax(predictions, dim=2)
            loss = compute_loss(predictions_scores, batch['target'].to(device), criterion)
            epoch_loss.append(loss.item())
            precision, recall, f1_score = monitor_model_eval(predictions_scores, batch)
            evaluation_measure['recall'].append(recall)
            evaluation_measure['precision'].append(precision)
            evaluation_measure['F1 score'].append(f1_score)
    if batch is not None:
        monitor_output_example(batch, dataloader, labels, predictions_scores)
    logger.info('model evaluation step is finished')
    return mean(epoch_loss), mean(evaluation_measure['recall']), \
           mean(evaluation_measure['precision']), mean(evaluation_measure['F1 score'])


@click.command()
@click.option('--learning-rate', required=True)
@click.option('--batch-size', required=True)
@click.option('--num-epochs', required=True)
@click.option('--folds', required=True)
@click.option('--model-name', required=True)
@click.option('--model-path', required=True)
@click.option('--dataset-path', required=True)
def main(learning_rate,
         batch_size,
         num_epochs,
         folds,
         model_name,
         model_path,
         dataset_path):
    hyperparameter_defaults = dict(
        lr=float(learning_rate),
        batch_size=int(batch_size),
        num_epochs=int(num_epochs),
        folds=int(folds),
        model_name=model_name,
        model_path=model_path,
    )
    wandb.init(project="360-ner-task", config=hyperparameter_defaults)
    device = torch.device('cpu' if torch.cuda.is_available() else 'cpu')
    dataset = sorted(glob.glob(dataset_path + '/*.pickle'), key=os.path.getsize)
    labels = json.load(open(glob.glob(dataset_path + '/*.json')[0], 'r'))['labels']
    num_labels = len(labels)
    logger.info("Using device: {}".format(device))
    logger.info("Using tokenizer from model : {}".format(model_name))
    tokenizer = init_tokenizer(hyperparameter_defaults['model_name'])
    criterion = nn.NLLLoss()
    folds = KFold(n_splits=hyperparameter_defaults['folds'], shuffle=False)
    for id_fold, fold in enumerate(folds.split(dataset)):
        logger.info('beginning fold n°{}'.format(id_fold + 1))
        model = init_model(hyperparameter_defaults['model_name'], device)
        classifier = classifier_factory('linear', device, **{'zn_size': model.config.hidden_size,
                                                             'num_labels': num_labels})
        logger.info(f'using the following classifier configuration : {classifier}')
        parameters = list(model.parameters()) + list(classifier.parameters())
        optimizer = optim.Adam(parameters, lr=hyperparameter_defaults['lr'])
        eval_dataloader, train_dataloader = create_dataloader(dataset, fold,
                                                              hyperparameter_defaults, tokenizer)
        wandb.watch(model)
        train_model(model, classifier, train_dataloader, optimizer, criterion, device, id_fold,
                    hyperparameter_defaults['num_epochs'])
        save_model(classifier, hyperparameter_defaults, id_fold, model, model_name)
        cv_loss, recall, precision, f1_score = eval_model(model, classifier,
                                                          eval_dataloader, criterion,
                                                          device, labels)
        # delete model to empty some space memory
        del model
        gc.collect()
        torch.cuda.empty_cache()

        # Monitor current fold
        wandb.log({"CV loss": cv_loss})
        wandb.log({"CV recall": recall})
        wandb.log({"CV precision": precision})
        wandb.log({"CV F1 score": f1_score})
        logger.info('🔎 evaluation is finished, here is the different metrics :')
        logger.info(f'cv recall : {recall}')
        logger.info(f'cv precision: {precision}')
        logger.info(f'cv F1 score: {f1_score}')
        logger.info(f'cv loss: {cv_loss}')


def create_dataloader(dataset, fold, hyperparameter_defaults, tokenizer):
    train_fold, eval_fold = fold
    train_fold = [dataset[file_index] for file_index in train_fold]
    eval_fold = [dataset[file_index] for file_index in eval_fold]
    train_dataset = CamemBertNerDataset(train_fold, tokenizer)
    eval_dataset = CamemBertNerDataset(eval_fold, tokenizer)
    train_dataloader = DataLoader(dataset=train_dataset,
                                  batch_size=hyperparameter_defaults["batch_size"],
                                  shuffle=False, drop_last=True,
                                  collate_fn=train_dataset.masked_lm_collate)
    eval_dataloader = DataLoader(dataset=eval_dataset,
                                 batch_size=hyperparameter_defaults["batch_size"],
                                 shuffle=False, drop_last=True,
                                 collate_fn=eval_dataset.masked_lm_collate)
    return eval_dataloader, train_dataloader


if __name__ == '__main__':
    # not used in this stub but often useful for finding various files
    project_dir = Path(__file__).resolve().parents[2]
    # find .env automagically by walking up directories until it's found, then
    # load up the .env entries as environment variables
    load_dotenv(find_dotenv())
    # pylint: disable=no-value-for-parameter
    main()



import logging
import os
from datetime import datetime
from os import path
from logging.config import fileConfig
import numpy as np
import wandb
from sklearn.metrics import precision_recall_fscore_support
from torch import nn
from transformers import AutoTokenizer, AutoModel
import torch

fileConfig(path.join(path.dirname(__file__), '../resources/log_config.ini'))
logger = logging.getLogger('camembert-ner-ft')


def count_parameters(mdl):
    return sum(p.numel() for p in mdl.parameters() if p.requires_grad)


def init_tokenizer(model_name):
    logger.info("🍌 Loading tokenizer...")
    tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast=True)
    return tokenizer


def mask_tokens(inputs, tokenizer):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    inputs = inputs.to(device)
    labels = inputs.clone().to(device)
    probability_matrix = torch.full(labels.shape, 0.15).to(device)
    masked_indices = torch.bernoulli(probability_matrix).bool().to(device)
    labels[~masked_indices] = -100  # We only compute loss on masked tokens

    # 80% of the time, we replace masked input tokens with tokenizer.mask_token ([MASK])
    indices_replaced = torch.bernoulli(torch.full(labels.shape, 0.8)).bool().to(device) \
                       & masked_indices
    inputs[indices_replaced] = tokenizer.get_vocab()['<mask>']

    # 10% of the time, we replace masked input tokens with random word
    indices_random = (torch.bernoulli(torch.full(labels.shape, 0.1)).bool().to(device)
                      & masked_indices & ~indices_replaced).to(device)
    random_words = torch.randint(len(tokenizer.get_vocab()) - 1,
                                 labels.shape, dtype=torch.long).to(device)
    inputs[indices_random] = random_words[indices_random]

    # The rest of the time (10% of the time) we keep the masked input tokens unchanged
    return inputs.tolist(), labels.tolist()


def save_model(classifier, hyperparameter_defaults, id_fold, model, model_name):
    logger.info("💾 saving model ..")
    current_datetime = datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
    save_location = hyperparameter_defaults['model_path']
    model_name = f'{model_name}-' \
                 f'{current_datetime}-fold-{id_fold + 1}.bin'
    if not os.path.exists(save_location):
        os.makedirs(save_location)
    save_location = os.path.join(save_location, model_name)
    torch.save({model, classifier}, save_location)
    wandb.save(save_location)


def monitor_output_example(batch, dataloader, labels, predictions_scores):
    decode_ids = dataloader.dataset.tokenizer.convert_ids_to_tokens(batch['tokens'][-1])
    seq_token_index = decode_ids.index('</s>')
    decode_ids = decode_ids[1:seq_token_index]
    true_labels = [labels[label_id] for label_id in batch['target'][-1][1:seq_token_index]]
    pred_labels = [labels[label_id]
                   for label_id in predictions_scores[-1].argmax(dim=1)[1:seq_token_index]]
    logger.info('here is one example tagged by the model (text/true labels/pred labels):')
    logger.info(decode_ids)
    logger.info(true_labels)
    logger.info(pred_labels)


def init_model(model_name, device):
    logger.info("🍌 Loading model...")
    model = AutoModel.from_pretrained(model_name)
    if device == 'cuda':
        model = nn.DataParallel(model).to(device)
    model = model.to(device)
    logger.info(f'The model has {count_parameters(model):,} trainable parameters')
    return model


def monitor_model_eval(predictions_scores: torch.Tensor,
                       batch: dict) -> (np.float, np.float, np.float):
    y_true = batch['target'].contiguous().reshape(-1)
    y_pred = predictions_scores.contiguous().argmax(dim=2).reshape(-1)
    pad_idx = (y_true == -100).nonzero(as_tuple=False).flatten()
    y_pred[pad_idx] = -100
    labels = np.intersect1d(y_pred.unique(), y_true.unique())
    labels = np.delete(labels, np.where(labels == -100))
    precision, recall, f1_score, _ = precision_recall_fscore_support(y_true,
                                                                     y_pred,
                                                                     average='macro',
                                                                     labels=labels)
    return precision, recall, f1_score



import logging
from logging.config import fileConfig
from os import path
from torch import nn

fileConfig(path.join(path.dirname(__file__), '../resources/log_config.ini'))
logger = logging.getLogger(__name__)


class CNN(nn.Module):
    def __init__(self, zn_size, num_labels):
        super().__init__()
        self.conv1d = nn.Conv1d(zn_size, num_labels, kernel_size=3, padding=1)

    def forward(self, z_n):
        return self.conv1d(z_n)


class LinearClassifier(nn.Module):
    def __init__(self, zn_size, num_labels):
        super().__init__()
        self.linear = nn.Linear(zn_size, num_labels)

    def forward(self, z_n):
        return self.linear(z_n)


def classifier_factory(model_name: str, device: str, **parameters) -> nn.Module:
    if model_name == 'linear':
        return LinearClassifier(parameters['zn_size'], parameters['num_labels']).to(device)
    raise ValueError(logger.error(f'{model_name}: no such model exists !'))



from setuptools import find_packages, setup

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

print(find_packages())
setup(
    name='name-entity-extraction-for-contextual-embedding',
    packages=find_packages(),
    version='0.4.0',
    description='a machine learning project for mlm task for contextual embedding',
    author='Simon Meoni',
    license='MIT',
    install_requires=requirements,
    entry_points='''
    [console_scripts]
    camembert_ner_ft=src.models.camembert_ner_ft:main
    make_dataset=src.data.make_dataset:main
    ''',
    package_data={
        "src": [
            "resources/log_config.ini",
        ]},
)
